from typing import Any
import asyncio
from aiogram.client.session import aiohttp
from rango_sdk.rango_response_entities import TransactionObject, MetaResponse, BalanceResponse, BestRouteResponse, \
    CreateTransactionResponse
from rango_sdk import config


class RangoClient:

    def __init__(self):
        super().__init__()
        self.api_key = config.RANGO_API_KEY
        self.base_url = config.RANGO_BASE_URL
        self.__meta = None
        self._popular_tokens = None
        asyncio.run(self.post_init())

    async def post_init(self):
        url = f"meta"
        response_json: MetaResponse = await self.__request(url, "GET")
        response_obj: MetaResponse = MetaResponse.from_dict(response_json)
        self.__meta = response_obj
        params = {'excludeNonPopulars': True}
        popular_response_json: MetaResponse = await self.__request(url, "GET", extra_params=params)
        popular_response: MetaResponse = MetaResponse.from_dict(popular_response_json)
        self._popular_tokens = popular_response
        print('Meta has been initialized...')

    def __get_token_data(self, blockchain: str, token: str):
        meta: MetaResponse = self.__meta
        for token in meta.tokens:
            if token.blockchain == blockchain.upper() and token.symbol == token:
                return token
            elif token.blockchain == blockchain.upper() and token.address == token.lower():
                return token

    async def get_meta(self) -> MetaResponse:
        return self.__meta

    async def popular_tokens(self):
        return self._popular_tokens['popularTokens']

    async def __request(self, url: str, method: str, data=None, extra_params=None, list_params=None) -> Any:
        if extra_params is None:
            extra_params = {}
        if data is None:
            data = {}

        params = {
            'apiKey': self.api_key
        }
        params.update(extra_params)
        encoded_params = '&'.join([f"{key}={value}" for key, value in params.items()])
        if list_params:
            encoded_params += '&'
            encoded_params += '&'.join([param for param in list_params])
        base_url = self.base_url + url
        req_url = base_url + '?' + encoded_params
        headers = {"accept": "*/*", "content-type": "application/json"}
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.request(
                    method,
                    headers=headers,
                    url=req_url,
                    json=data
            ) as resp:
                response = await resp.json()
                return response

    async def balance(self, wallet_addresses: list[str]) -> BalanceResponse:
        url = f"wallets/details"
        list_params = []
        for bwa in wallet_addresses:
            list_params.append(f'address={bwa}')
        response_json: dict = await self.__request(url, "GET", list_params=list_params)
        balance_response: BalanceResponse = BalanceResponse.from_dict(response_json)
        return balance_response

    async def route(self, _connected_wallets: list, selected_wallets: dict, from_blockchain: str,
                    from_token_identifier: str,
                    to_blockchain: str, to_token_identifier: str, amount: float) -> BestRouteResponse:
        url = f"routing/best"
        from_token = self.__get_token_data(from_blockchain, from_token_identifier)
        to_token = self.__get_token_data(to_blockchain, to_token_identifier)
        connected_wallets = []
        for item in _connected_wallets:
            blockchain, address = item.split('.')
            connected_wallets.append(
                {'blockchain': blockchain, 'addresses': [address]}
            )
        payload = {
            "from": {
                "blockchain": from_blockchain.upper(),
                "symbol": from_token.symbol,
                "address": from_token.address
            },
            "to": {
                "blockchain": to_blockchain.upper(),
                "symbol": to_token.symbol,
                "address": to_token.address
            },
            "checkPrerequisites": False,
            "connectedWallets": connected_wallets,
            "selectedWallets": selected_wallets,
            "amount": amount,
            "maxLength": 1
        }
        response_json: dict = await self.__request(url, "POST", data=payload)
        best_route: BestRouteResponse = BestRouteResponse.from_dict(response_json)
        return best_route

    async def create_transaction(self, request_id: str, step: int = 1, slippage: int = 2) -> CreateTransactionResponse:
        url = f"tx/create"
        payload = {
            "userSettings": {
                "slippage": slippage
            },
            "validations": {
                "balance": True,
                "fee": True,
                "approve": True
            },
            "requestId": request_id,
            "step": step,
        }
        response_json: dict = await self.__request(url, "POST", data=payload)
        create_tx: CreateTransactionResponse = CreateTransactionResponse.from_dict(response_json)
        return create_tx

    async def check_approval(self, request_id: str) -> bool:
        url = f"tx/{request_id}/check-approval"
        response: dict = await self.__request(url, "GET")
        print(response)
        return response["isApproved"]

    async def check_tx(self, request_id: str, tx_id: str, step: int) -> TransactionObject:
        url = f"tx/check-status"
        payload = {
            "requestId": request_id,
            "txId": tx_id,
            "step": step,
        }
        response_json = await self.__request(url, "POST", data=payload)
        transaction_object: TransactionObject = TransactionObject.from_dict(response_json)
        return transaction_object




from rango_sdk.rango_client import RangoClient
from rango_sdk.rango_response_entities import BestRouteResponse, CosmosTransaction, EvmTransaction, SolanaTransaction, \
    StarkNetTransaction, TransferTransaction, TrxTransaction

__all__ = ['RangoClient']



from os import environ as env
from dotenv import load_dotenv
load_dotenv()

TOKEN = env['TOKEN']
RANGO_API_KEY = env['RANGO_API_KEY']
RANGO_BASE_URL = env['RANGO_BASE_URL']
WEBHOOK_URL = env['WEBHOOK_URL']


from dataclasses import dataclass, field
from typing import Optional, List, Dict, Union
from dataclasses_json import dataclass_json, config
from decimal import Decimal


@dataclass_json
@dataclass
class ExplorerUrl:
    url: str
    description: str


@dataclass_json
@dataclass
class BridgeExtra:
    requireRefundAction: bool
    srcTx: str
    destTx: str


@dataclass_json
@dataclass
class TransactionObject:
    status: str
    timestamp: int
    outputAmount: str
    explorerUrl: List[ExplorerUrl]
    bridgeExtra: BridgeExtra
    extraMessage: Optional[str] = None
    failedType: Optional[str] = None
    referrals: Optional[str] = None
    newTx: Optional[str] = None
    diagnosisUrl: Optional[str] = None
    steps: Optional[str] = None
    outputToken: Optional[str] = None

    def is_successful(self) -> bool:
        return self.status == "success"

    def get_output_amount(self) -> str:
        return '%.3f' % Decimal(self.outputAmount)

    def print_explorer_urls(self):
        msg = ''
        for ex in self.explorerUrl:
            msg += f'🔹 [Explorer Link]({ex.url}) -> {ex.description} \n'
            return msg


@dataclass_json
@dataclass
class Amount:
    amount: int
    decimals: int


@dataclass_json
@dataclass
class Asset:
    blockchain: str
    symbol: str
    address: Optional[str] = None
    name: Optional[str] = None
    decimals: Optional[int] = None

    def __repr__(self):
        return f'{self.blockchain}.{self.symbol}'


@dataclass_json
@dataclass
class SwapFee:
    asset: Asset
    expenseType: str
    amount: float
    name: str


@dataclass_json
@dataclass
class SwapNode:
    marketName: str
    percent: float
    marketId: Optional[str] = None


@dataclass_json
@dataclass
class SwapNode:
    nodes: List[SwapNode]


@dataclass_json
@dataclass
class SwapRoute:
    nodes: List[SwapNode]


@dataclass_json
@dataclass
class SwapResultAsset:
    symbol: str
    logo: str
    blockchainLogo: str
    blockchain: str
    decimals: int
    usdPrice: Optional[float] = None
    address: Optional[str] = None


@dataclass_json
@dataclass
class RecommendedSlippage:
    error: bool
    slippage: Optional[float] = None


@dataclass_json
@dataclass
class SwapResult:
    swapperId: str
    swapperType: str
    from_: SwapResultAsset = field(metadata=config(field_name='from'))
    to: SwapResultAsset
    fromAmount: float
    toAmount: float
    fee: List[SwapFee]
    estimatedTimeInSeconds: int
    swapChainType: str
    fromAsset: Optional[Asset] = None
    toAsset: Optional[Asset] = None
    swapperLogo: Optional[str] = None
    fromAmountPrecision: Optional[float] = None
    fromAmountMinValue: Optional[float] = None
    fromAmountMaxValue: Optional[float] = None
    fromAmountRestrictionType: Optional[str] = None
    routes: Optional[List[SwapRoute]] = None
    recommendedSlippage: Optional[RecommendedSlippage] = None
    warnings: Optional[List[str]] = None
    timeStat: Optional[Dict[str, int]] = None
    includesDestinationTx: Optional[bool] = None
    maxRequiredSign: Optional[int] = None
    isWrapped: Optional[bool] = None


@dataclass_json
@dataclass
class SimulationResult:
    outputAmount: float
    swaps: List[SwapResult]
    resultType: str


@dataclass_json
@dataclass
class BestRouteResponse:
    requestAmount: float
    requestId: str
    result: SimulationResult
    walletNotSupportingFromBlockchain: bool
    processingLimitReached: bool
    missingBlockchains: List[str]
    diagnosisMessages: List[str]
    compareStatus: str
    from_: Asset = field(metadata=config(field_name='from'))
    to: Asset


@dataclass_json
@dataclass
class Token:
    blockchain: str
    symbol: str
    isPopular: bool
    supportedSwappers: List[str]
    image: Optional[str] = None
    address: Optional[str] = None
    usdPrice: Optional[float] = None
    decimals: Optional[int] = None
    name: Optional[str] = None
    isSecondaryCoin: Optional[bool] = None
    coinSource: Optional[str] = None
    coinSourceUrl: Optional[str] = None


@dataclass_json
@dataclass
class BlockchainMeta:
    name: str
    defaultDecimals: int
    addressPatterns: List[str]
    feeAssets: List[Asset]
    logo: str
    displayName: str
    shortName: str
    sort: int
    color: str
    enabled: bool
    type: str
    chainId: Optional[str] = None
    info: Optional[dict] = None


@dataclass_json
@dataclass
class SwapperMetaDto:
    id: str
    title: str
    logo: str
    swapperGroup: str
    types: List[str]
    enabled: bool


@dataclass_json
@dataclass
class MetaResponse:
    tokens: List[Token]
    popularTokens: List[Token]
    blockchains: List[BlockchainMeta]
    swappers: List[SwapperMetaDto]


@dataclass_json
@dataclass
class GetWalletDetailsRequest:
    addresses: List[str]


@dataclass_json
@dataclass
class Balance:
    asset: Asset
    amount: Amount


@dataclass_json
@dataclass
class WalletDetails:
    address: str
    blockChain: str
    balances: List[Balance]


@dataclass_json
@dataclass
class BalanceResponse:
    wallets: List[WalletDetails]


# Create Transaction
@dataclass_json
@dataclass
class GenericTransaction:
    type: str


@dataclass_json
@dataclass
class CosmosTransaction(GenericTransaction):
    fromWalletAddress: str
    blockChain: str
    data: dict
    rawTransfer: Optional[dict] = None


@dataclass_json
@dataclass
class EvmTransaction(GenericTransaction):
    blockChain: str
    isApprovalTx: bool
    to: str
    data: Optional[str] = None
    value: Optional[str] = None
    gasLimit: Optional[str] = None
    gasPrice: Optional[int] = None
    maxPriorityFeePerGas: Optional[int] = None
    maxFeePerGas: Optional[int] = None
    nonce: Optional[int] = None
    from_: Optional[str] = field(default=None, metadata={"name": "from"})


@dataclass_json
@dataclass
class SolanaTransaction(GenericTransaction):
    blockChain: str
    from_: str = field(metadata={"name": "from"})
    identifier: str
    instructions: List[dict]
    signatures: List[dict]
    txType: str
    serializedMessage: Optional[str] = None
    recentBlockhash: Optional[str] = None


@dataclass_json
@dataclass
class StarkNetTransaction(GenericTransaction):
    blockChain: str
    calls: List[dict]
    isApprovalTx: bool
    maxFee: Optional[int] = None


@dataclass_json
@dataclass
class TransferTransaction(GenericTransaction):
    method: str
    fromWalletAddress: str
    recipientAddress: str
    amount: int
    decimals: int
    asset: Asset
    memo: Optional[str] = None


@dataclass_json
@dataclass
class TrxTransaction(GenericTransaction):
    blockChain: str
    isApprovalTx: bool
    txID: str
    visible: bool
    raw_data: Optional[dict] = None
    raw_data_hex: Optional[str] = None
    externalTxId: Optional[str] = None
    __payload__: Optional[dict] = None


@dataclass_json
@dataclass
class CreateTransactionResponse:
    ok: bool
    transaction: Union[
        CosmosTransaction, EvmTransaction, SolanaTransaction, StarkNetTransaction, TransferTransaction, TrxTransaction]
    error: Optional[str] = None
    errorCode: Optional[int] = None



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name="rango-sdk",
    version="0.4.0",
    packages=find_packages(),
    install_requires=requirements,
    long_description=open('README.md').read(),
    author='Nimara0x',
    author_email='nimcrypto12@gmail.com',
    url='https://github.com/Nimara0x/rango-sdk',
    python_requires='>=3.9',
)